/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Sept 2001      */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 

// initial positions for iconbars
int nexticon_x=0, nexticon_y=22;
Display *dpy; 
Window root, check_win, lastenter;
Window lastfocus=0; 
int hacking_focus=False, inhibit_errors=False;
Cursor resize_curs, move_curs, normal_curs; 
int screen, iconbar_state=ICONBAR_HIDDEN; 
Client *overicon=NULL;
Client *client_list=NULL; 
long client_count; 
Config config; 
Atom icccm[ICCCM_ATOM_COUNT], motif[MOTIF_ATOM_COUNT], gnome[GNOME_ATOM_COUNT]; 
char *rcfile=RCFILE;

#ifdef SHAPE_SUPPORT 
int shape, shape_event; 
#endif 
 
#ifdef WORKSPACE_SUPPORT 
int workspace=0; 
#endif 
 
#ifdef XINERAMA_SUPPORT 
int xinerama; 
int xinerama_heads;
#ifndef EMULATE_XINERAMA
XineramaScreenInfo *xinerama_screens; 
#else
XineramaScreenInfo xinerama_screens[EMULATED_SCREENS]=
{
  {0, 0, 0, 600, 1003},
  {0, 600, 100, 400, 903},
  {0, 1000, 200, 280, 803} 
};
#endif
#endif 

#ifdef I18N_SUPPORT
int use_fontset=False;
#endif
 
/*  
 * 
 * Quit nicely 
 * 
 */ 

void quit() 
{ 
  sprintf(lastsub,"TRACE1: quit\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: quit\n"); 
#endif 
 
  Restart_XDarwin();
  
  /* Remove all frames */ 
  while(client_list)  
    delete_client(client_list, True); 
 
  /* Free cursors */ 
  XFreeCursor(dpy, normal_curs);
  XFreeCursor(dpy, resize_curs); 
  XFreeCursor(dpy, move_curs); 
 
  /* Free theme data */ 
  free_config(); 

  XSetInputFocus(dpy, root, RevertToPointerRoot, CurrentTime); 
  XCloseDisplay(dpy);
  if(gQuitFlag==0)
    gQuitFlag=1; 

  if(Display_Name) free(Display_Name);
} 

/* 
 * 
 * Handle signals 
 * 
 */ 
void signal_handler(int signal) 
{
  int dummy;
  
  sprintf(lastsub,"TRACE1: signal_handler: %d\n",signal); 
  subtrace();
//#ifdef SHOW_TRACE1 
  if(signal!=SIGCHLD) printf("TRACE1: signal_handler: %d\n",signal); 
//#endif 
 
  switch(signal) 
  { 
  case SIGINT: 
  case SIGTERM: 
    quit(); 
    break;
  case SIGHUP:
//    Restart_XDarwin();
//    reload_config();
    show_last_fifteen();
    dummy=sleep(1);
    //system("/bin/sleep 1");
//    printf("quitting...\n");
//    quit();
    break;
  case SIGCHLD: 
//    wait(NULL); 
    break; 
  } 
} 
 
/* 
 * 
 * Create a frame for each window that already exists 
 * 
 */ 

void scan_for_windows() 
{ 
  unsigned int count, i; 
  Window w1, w2, *windows=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE1: scan_for_windows\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: scan_for_windows\n"); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0; i<count; i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(!attr.override_redirect && attr.map_state==IsViewable) 
      create_new_client(windows[i], False); 
  } 
  if(windows) XFree(windows); 
} 
 
/* 
 * 
 * Initialize window manager 
 * 
 */ 

void init_wm() 
{ 
  int dummy;
  int i, head=0;
 
  sprintf(lastsub,"TRACE1: init_wm\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: init_wm\n"); 
#endif 
 
  /* ICCCM stuff */ 
  icccm[WM_STATE]=XInternAtom(dpy, "WM_STATE", False); 
  icccm[WM_PROTOCOLS]=XInternAtom(dpy, "WM_PROTOCOLS", False); 
  icccm[WM_DELETE]=XInternAtom(dpy, "WM_DELETE_WINDOW", False); 
  icccm[WM_CHANGE_STATE]=XInternAtom(dpy, "WM_CHANGE_STATE", False); 
 
  /* Mwm hints */ 
  motif[MOTIF_WM_HINTS]=XInternAtom(dpy, "_MOTIF_WM_HINTS", False); 
 
  /* GNOME hints */ 
  gnome[WIN_CLIENT_LIST]=XInternAtom(dpy, "_WIN_CLIENT_LIST", False); 
  gnome[WIN_STATE]=XInternAtom(dpy, "_WIN_STATE", False); 
  gnome[WIN_HINTS]=XInternAtom(dpy, "_WIN_HINTS", False); 
  gnome[WIN_LAYER]=XInternAtom(dpy, "_WIN_LAYER", False); 
  gnome[WIN_SUPPORTING_WM_CHECK]=XInternAtom(dpy, "_WIN_SUPPORTING_WM_CHECK", False); 
  gnome[WIN_DESKTOP_BUTTON_PROXY]=XInternAtom(dpy, "_WIN_DESKTOP_BUTTON_PROXY", False); 
#ifdef WORKSPACE_SUPPORT 
  gnome[WIN_WORKSPACE]=XInternAtom(dpy, "_WIN_WORKSPACE", False); 
  gnome[WIN_WORKSPACE_COUNT]=XInternAtom(dpy, "_WIN_WORKSPACE_COUNT", False); 
#endif 
 
  /* Create GNOME check window */ 
  check_win=XCreateSimpleWindow(dpy, root, -200, -200, 5, 5, 0, 0, 0); 
   
  /* Set up GNOME properties */ 
  set_gnome_hint(check_win, WIN_SUPPORTING_WM_CHECK, check_win); 
  set_gnome_hint(root, WIN_SUPPORTING_WM_CHECK, check_win); 
  set_gnome_hint(check_win, WIN_DESKTOP_BUTTON_PROXY, check_win); 
  set_gnome_hint(root, WIN_DESKTOP_BUTTON_PROXY, check_win); 
#ifdef WORKSPACE_SUPPORT 
  set_gnome_hint(root, WIN_WORKSPACE_COUNT, config.workspace_count); 
  set_gnome_hint(root, WIN_WORKSPACE, 0); 
#endif 
  
#ifdef SHAPE_SUPPORT 
  /* Check for shape support */ 
  shape=XShapeQueryExtension(dpy, &shape_event, &dummy); 
#endif 
 
  head=0;
#ifdef XINERAMA_SUPPORT
#ifndef EMULATE_XINERAMA 
  /* Check xinerama extension */ 
  xinerama=XineramaQueryExtension(dpy, &dummy, &dummy); 
  if(xinerama) {
    xinerama_screens=XineramaQueryScreens(dpy, &xinerama_heads);
    head=find_head(50,50);
  }
#else
  xinerama=True;
  xinerama_heads=EMULATED_SCREENS;
#endif
  if(xinerama) {
    printf("Xinerama screens found: %d\n",xinerama_heads);
    for(i=0;i<xinerama_heads;i++)
      printf("  head %d (x,y w,h):  %d,%d %d,%d\n",i,screen_x(i),screen_y(i),screen_width(i),screen_height(i));
  } else
    printf("Xinerama not found...\n");
#endif 
 
  /* Set event mask for root window */ 
  XSelectInput(dpy, root, SubstructureRedirectMask|SubstructureNotifyMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|PropertyChangeMask); 
 
  normal_curs=XCreateFontCursor(dpy, NORMAL_CURS);
  resize_curs=XCreateFontCursor(dpy, RESIZE_CURS); 
  move_curs=XCreateFontCursor(dpy, MOVE_CURS); 

// set up the correct iconbar positions, depending on iconbar location option
  if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_TOPRIGHT)
    nexticon_y=screen_y(head)+22;
  else
    nexticon_y=screen_y(head)+screen_height(head)-config.title.height;

  if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT)
    nexticon_x=screen_x(head);
  else
    nexticon_x=screen_x(head)+screen_width(head)-1;
} 
 
/* 
 * 
 * Set up signal handler 
 * 
 */ 

void init_signals() 
{ 
  struct sigaction act; 
 
  sprintf(lastsub,"TRACE1: init_signals\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: init_signals\n"); 
#endif 
  
  act.sa_handler=signal_handler; 
  act.sa_flags=0; 
  sigaction(SIGTERM, &act, NULL); 
  sigaction(SIGINT, &act, NULL); 
  sigaction(SIGHUP, &act, NULL); 
  sigaction(SIGCHLD, &act, NULL); 
} 
 
/* 
 * 
 * Main routine for 'oroborus part' - now exits back to main, rather than calling the event loop
 * 
 */ 

int orob_main(int argc, char *argv[]) 
{
  char *dpy_str=NULL; 
  int i;
  char def_dpy[10]=":0.0";
  
#ifdef I18N_SUPPORT 
  char *locale=NULL; 
#endif /* I18N */ 

  sprintf(lastsub,"TRACE1: orob_main (argc: %d)\n",argc); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: orob_main (argc: %d)\n",argc); 
#endif 

  on_cmd_line=1;
  for(i=0;i<argc;i++)
  {
#ifdef DEBUG
    printf("arg %d :  %s\n",i,argv[i]);
#endif
    if(!strncmp(argv[i], "-psn", 4))
      on_cmd_line=0;
    if(!strncmp(argv[i], "-config", 7) && i+1<argc)
      rcfile=argv[++i];
    else if(!strncmp(argv[i], "-display", 8) && i+1<argc)
      dpy_str=argv[++i];
  }
 
#ifdef I18N_SUPPORT 
  locale=setlocale(LC_ALL, "");
  if(locale!=NULL)
  {
#ifdef DEBUG
    printf("locale=%s\n", locale);
#endif
    if(strncmp(locale, "C", 1) && strncmp(locale, "POSIX", 5) && XSupportsLocale())
      use_fontset=True;
  }
#ifdef DEBUG
  if(use_fontset)
    printf("setting use_fontset=True\n");
  else
    printf("setting use_fontset=False\n");
#endif
#endif /* I18N */ 
   
  init_signals(); 

  /* Initialize X */ 
  if( dpy_str==NULL ) {
    if(!(dpy=XOpenDisplay(dpy_str)))
      dpy_str=def_dpy;
    printf("Display string: %s %#lx\n",dpy_str,dpy);
  }
  
  if(!(dpy=XOpenDisplay(dpy_str))) {
    if(on_cmd_line)
        fprintf(stderr, "%s: Can't open display, X may not be running.\n", PROGNAME);
    return -1;
  }
 
// Get the display name to send to the X11 app launcher
  if(dpy_str==NULL) {
    dpy_str=getenv("DISPLAY");
    printf("Display string: %s %#lx\n",dpy_str,dpy);
  }
  
  if(dpy_str!=NULL)
    Display_Name=strdup(dpy_str);
  else {
    fprintf(stderr,"OroborOSX: Could not find name of display\n");
    return -2;
  }
  
  XSetErrorHandler(handle_xerror); 
   
  root=XDefaultRootWindow(dpy); 
  screen=XDefaultScreen(dpy); 
 
  load_config(); 
  init_wm();
#ifdef KEYBOARD_SUPPORT
  init_keyboard();
#endif 
  scan_for_windows(); 

//#ifdef KEYBOARD_SUPPORT
  XMapWindow(dpy, check_win);
//  grab_keys(check_win); 
  set_focus_client(get_focus_client());
//#endif 

  XDefineCursor(dpy, root, normal_curs);

  return 0; 
}
